#!/usr/bin/env python3

# Ceci est un assembleur pour la micromachine.

# Pour que cela marche:      pip install lark bitarray
# Syntaxe: python mm_asm.py input [output] (l'output par defaut est a.hex) 
# Auteur: Laurent Turpin, longue vie a lui.

import sys
from lark import Lark, UnexpectedInput
from bitarray import bitarray
from bitarray.util import ba2hex, int2ba

grammar = """
start : bin_op
      | una_op
      | cmp
      | affect
      | mem_read
      | mem_write
      | abs_jump
      | rel_jump
      |

bin_op : expr ARROW reg

una_op : UN_OP reg ARROW reg
       | reg ARROW reg

cmp    : expr_minus CMP_END

affect : reg_cst ARROW reg

mem_read: pointer ARROW reg

mem_write: reg ARROW pointer

abs_jump: "JA" NUM

rel_jump: "JR" signed_num COND?

expr: expr_plus | expr_minus | expr_and | expr_or | expr_xor
expr_plus: reg "+" reg_cst
expr_minus: reg "-" reg_cst
expr_and: reg "and" reg_cst
expr_or: reg "or" reg_cst
expr_xor: reg "xor" reg_cst

reg : B | A
reg_cst: signed_num | A
signed_num: NUM | MINUS NUM

pointer: STAR (NUM | A)

A: "A"
B: "B"
MINUS : "-"
OP : "+" | "and" | "or" | "xor"
CMP_END : "?"
UN_OP : "lsr" | "not"
ARROW : "->"
STAR : "*"
COND : "IFZ" | "IFC" | "IFN"


COMMENT: /;.*/
%ignore COMMENT


%import common.NUMBER -> NUM
%import common.WS
%ignore WS
"""

parser = Lark(grammar)

codeop_dict = {'expr_plus':  0,
               'expr_minus': 1,
               'expr_and':   2,
               'expr_or':    3,
               'expr_xor':   4,
               'lsr':        5,
               'cmp':        6,
               'not':        8,
               'affect':     9,
               'mem_read':  13,
               'mem_write': 14,
               'abs_jump':  15
               }

cond_dict = {'IFZ': 1,
             'IFC': 2,
             'IFN': 3,
             }


def get_cst(sub_tree) -> int:
    cst = int(sub_tree.children[-1])
    if len(sub_tree.children) > 1:
        # A minus sign is used
        cst = cst * -1
    return cst


def ast_to_hex(tree) -> str:
    instruction = tree.children[0]
    operation_type = instruction.data

    out_string: str

    if operation_type == 'rel_jump':
        msb = 1
        cond = 0
        offset = get_cst(instruction.children[0])
        if len(instruction.children) > 1:
            cond = cond_dict[instruction.children[1].value]
        bin_instr = bitarray()
        bin_instr.append(msb)
        bin_instr.extend(int2ba(cond, 2))
        bin_instr.extend(int2ba(offset, 5, signed=True))
        out_string = ba2hex(bin_instr)
        out_string += '\t# ' + bin_instr.to01()
    else:
        msb = 0
        codeop: int
        arg1s: int
        arg2s: int
        dests: int
        use_cst = False
        signed_cst: bool
        cst: int

        if operation_type == 'bin_op':
            expr = instruction.children[0].children[0]
            dest = instruction.children[2].children[0]
            # determine codeop
            codeop = codeop_dict[expr.data]
            dests = int(dest == 'B')
            expr = expr.children
            arg1s = int(expr[0].children[0] == 'B')
            arg2s = int(expr[1].children[0] != 'A')
            # Use a constant
            if arg2s == 1:
                use_cst = True
                signed_cst = True
                cst = get_cst(expr[1].children[0])

        elif operation_type == 'una_op':
            dest = instruction.children[-1].children[0]
            arg1 = instruction.children[-3].children[0]
            if len(instruction.children) == 4:  # 'lsr' or 'not'
                codeop = codeop_dict[instruction.children[0].value]
                arg2s = 1
            else:  # simple copy
                codeop = codeop_dict['not']
                arg2s = 0
            arg1s = int(arg1 == 'B')
            dests = int(dest == 'B')
        elif operation_type == 'cmp':
            codeop = codeop_dict[operation_type]
            dests = 0
            expr = instruction.children[0].children
            arg1s = int(expr[0].children[0] == 'B')
            arg2s = int(expr[1].children[0] != 'A')
            # Use a constant
            if arg2s == 1:
                use_cst = True
                signed_cst = True
                cst = get_cst(expr[1].children[0])
        elif operation_type == 'affect':
            codeop = codeop_dict[operation_type]
            arg1s = 0
            arg2 = instruction.children[0].children[0]
            dest = instruction.children[-1].children[0]

            dests = int(dest == 'B')
            arg2s = int(arg2 != 'A')
            # Use a constant
            if arg2s == 1:
                use_cst = True
                signed_cst = True
                cst = get_cst(arg2)
        elif operation_type == 'mem_read':
            codeop = codeop_dict[operation_type]
            arg1s = 0
            dest = instruction.children[-1].children[0]
            arg2 = instruction.children[0].children[-1]

            dests = int(dest == 'B')
            arg2s = int(arg2 != 'A')
            # Use a constant
            if arg2s == 1:
                use_cst = True
                signed_cst = False
                cst = int(arg2)
        elif operation_type == 'mem_write':
            codeop = codeop_dict[operation_type]
            dests = 0
            arg1 = instruction.children[0].children[0]
            arg2 = instruction.children[-1].children[-1]

            arg1s = int(arg1 == 'B')
            arg2s = int(arg2 != 'A')
            # Use a constant
            if arg2s == 1:
                use_cst = True
                signed_cst = False
                cst = int(arg2)
        else:  # operation_type == 'abs_jump'
            codeop = codeop_dict[operation_type]
            arg1s = 0
            arg2s = 0
            dests = 0
            use_cst = True
            signed_cst = False
            cst = int(instruction.children[0])

        bin_instr = bitarray()
        bin_instr.append(msb)
        bin_instr.extend(int2ba(codeop, 4))
        bin_instr.append(arg2s)
        bin_instr.append(arg1s)
        bin_instr.append(dests)
        out_string = ba2hex(bin_instr)
        out_string += '\t# ' + bin_instr.to01()

        if use_cst:
            bin_cst = int2ba(cst, length=8, signed=signed_cst)
            out_string += '\n' + ba2hex(bin_cst)
            out_string += '\t# ' + bin_cst.to01()
    return out_string


def asm_to_hex(instr: str) -> str:
    ast = parser.parse(instr)
    if len(ast.children) == 0:
        return ''
    return ast_to_hex(ast)


def file_asm2file_hex(infile_str, outfile_str):
    hex_str = ''
    line_nb = 1
    error = False
    try:
        infile = open(infile_str, 'r')
    except (OSError, IOError):
        print('Error: {}: cannot read input file'.format(infile_str), file=sys.stderr)
        exit(1)

    for line in infile:
        hex_line = ''
        try:
            hex_line = asm_to_hex(line)
            if hex_line != '':
                hex_line += ',\t' + line
        except (UnexpectedInput, OverflowError):
            error = True
            print('Error in line {}: {}'.format(line_nb, line),
                  file=sys.stderr)
        hex_str += hex_line
        line_nb += 1
    if error or hex_str == '':
        print('No Output file')
    else:
        try:
            outfile = open(outfile_str, 'w')
        except (OSError, IOError):
            print('Error: {}: cannot write to output file'.format(outfile_str), file=sys.stderr)
            exit(1)
        outfile.write('v2.0 raw\n')
        outfile.write(hex_str)
        outfile.write('\n')
        print('Output written to {}'.format(outfile_str))


def print_usage():
    print('Usage: python {} INPUT_FILE [OUTPUT_FILE]'.format(sys.argv[0]))
    print('')
    print('OUTPUT_FILE is optional, its default value is a.hex')


if __name__ == '__main__':
    args = sys.argv[1:]
    if len(args) < 1 or len(args) > 2:
        print_usage()
    else:
        infile_str = args[0]
        if len(args) == 2:
            outfile_str = args[1]
        else:
            outfile_str = 'a.hex'
        file_asm2file_hex(infile_str, outfile_str)

